#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : estpsp.py
# Contents: Estimate a pattern spectrum from data
# Author  : Christian Borgelt
# History : 2013.12.18 file created
#           2013.12.19 second version of slot counting added
#           2013.12.20 logarithmic computation of expected values
#           2014.01.10 slot counting improved, other versions removed
#           2014.01.13 interfaces of estimation functions changed
#           2014.02.21 more sophisticated estimation method added
#           2014.03.28 slot counting improved (sentinel, dictionary)
#           2014.03.06 interfaces of functions simplified (defaults)
#           2014.07.25 spectrum estimation for item sequences added
#           2014.10.12 adapted to modified interface of estpsp()
#-----------------------------------------------------------------------
from sys    import argv, stderr, version_info
from time   import time
from math   import floor, ceil, sqrt, exp, log, lgamma
from random import random, sample, shuffle
from random import seed as srand

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# File Reading Function
#-----------------------------------------------------------------------

def getrec (f, recseps, fldseps, blanks, comment):
    '''Read a record from a file and split it into fields.
f       file to read from
recseps record  separators
fldseps field   separators
blanks  blank   characters
comment comment characters
returns a list of the fields read or None at end of file'''
    c = f.read(1).decode()      # read the next character
    while c != '' and c in comment:
        while c != '' and c not in recseps:
            c = f.read(1).decode() # skip comment until record separator
        c = f.read(1).decode()  # consume the record separator
    if c == '': return None     # if at end of file, abort
    rec = []                    # initialize the record
    sep = fldseps +recseps      # combine all separators
    while 1:                    # field read loop
        while c != '' and c in blanks:
            c = f.read(1).decode() # skip leading blanks
        fld = ''                # initialize the field
        while c != '' and c not in sep:
            fld += c            # append character to the field
            c = f.read(1).decode()  # and read the next character
        fld = fld.strip(blanks) # remove leading/trailing blanks */
        if len(fld) > 0:        # if the field is not empty,
            rec.append(fld)     # append the field to the record
        if c == '' or c in recseps: break
        c = f.read(1).decode()  # consume field separator
    return rec                  # return the record

#-----------------------------------------------------------------------
# Pattern Spectrum Estimation Functions
#-----------------------------------------------------------------------

def factorial (n):
    '''factorial (n)
Compute the factorial of n, that is, n!
n       number of which to compute the factorial
returns the factorial of n, that is, n!'''
    if n <  0: return 0         # treat special cases
    if n <= 1: return 1         # 0! = 1! = 1
    r = 1                       # initialize the product
    for i in range(1,n): r *= i # multiply with sequence of numbers
    return r                    # return the result

#-----------------------------------------------------------------------

def choose (n, k):
    '''choose (n, k)
Compute the binomial coefficient n choose k.
n       number of objects to choose from
k       number of objects to choose
returns the number of possibilities to choose k of n objects'''
    if k < 0 or k > n: return 0 # check for valid parameters
    if n-k < k: k = n-k         # exploit (n choose k) = (n choose n-k)
    r = 1                       # initialize with (n choose 0)
    for i in range(1,k+1):      # traverse the factors
        r = (r *n) /i           # multiply with the next factor
        n -= 1                  # (all intermediate results are integer)
    return r                    # return computed binomial coefficient

#-----------------------------------------------------------------------

def countslotsx (trains, width):
    '''countslotsx (trains, width)
Count slots for item sets per item set size.
trains  (spike) train database to analyze
        The database must be an iterable of (spike) trains;
        each train must be a pair, consisting of a item/neuron
        identifier and an iterable of points/spike times;
        each item/neuron identifier must be a hashable object.
width   window width/maximal distance between synchronous spikes
returns a list of slot counters per item set size'''
    points = [(s,i) for i,t in trains for s in t]
    if not points: return []    # create labeled point process
    points.sort()               # sort the points/spike times
    points += [(oo,-1)]         # add a sentinel at the end
    cnts = [0 for p in points]+[0]
    occs = dict()               # set up a slot counter array
    r    = 0                    # and an item occurrence dictionary
    for l in range(len(points)-1): # traverse the points/spikes
        e = points[l][0] +width # compute the end of the window
        while points[r][0] <= e:# while point is in current window
            k = points[r][1]    # update item occurrence dictionary
            if k in occs: occs[k] += 1
            else:         occs[k]  = 1
            r += 1              # go to the next point
        cnts[len(occs)] += 1    # count for number of distinct items
        i = points[l][1]        # remove occurrence of first point
        if occs[i] <= 1: del occs[i]
        else:            occs[i] -= 1
    for i in reversed(range(len(cnts))):
        if cnts[i] > 0: break   # reduce to populated counters
    cnts = cnts[0:i+1]          # (trim zero entries at the end)
    for i in range(1,len(cnts)):# traverse the item set sizes
        for k in range(1,i):    # considering all possible subsets
            cnts[k] += cnts[i] *choose(i-1,k-1)
    return cnts                 # return list of slot counters per size

#-----------------------------------------------------------------------

def countslots (trains, width):
    '''countslots (trains, width)
Count slots for item sets per item set size.
trains  (spike) train database to analyze
        The database must be an iterable of (spike) trains;
        each train must be a pair, consisting of a item/neuron
        identifier and an iterable of points/spike times;
        each item/neuron identifier must be a hashable object.
width   window width/maximal distance between synchronous spikes
returns a list of slot counters per item set size'''
    points = [(s,i) for i,t in trains for s in t]
    if not points: return []    # create labeled point process
    points.sort()               # sort the points/spike times
    points += [(oo,-1)]         # add a sentinel at the end
    cnts = [0 for p in points]+[0]
    r    = 0                    # set up a slot counter array
    for l in range(len(points)-1): # traverse the points/spikes
        e = points[l][0] +width # compute the end of the window
        while points[r][0] <= e:# while point is in current window
            r += 1              # go to the next point
        cnts[r-l] += 1          # count for number of distinct items
    for i in reversed(range(len(cnts))):
        if cnts[i] > 0: break   # reduce to populated counters
    cnts = cnts[0:i+1]          # (trim zero entries at the end)
    for i in range(1,len(cnts)):# traverse the item set sizes
        for k in range(1,i):    # considering all possible subsets
            cnts[k] += cnts[i] *choose(i-1,k-1)
    return cnts                 # return list of slot counters per size

#-----------------------------------------------------------------------

def estpsp_dir (trains, target='s', width=0.003, supp=2,
                zmin=2, zmax=None, report='#', equiv=10000):
    '''estpsp_dir (trains, target='s', width=0.003, supp=2,
            zmin=2, zmax=None, report='#', equiv=10000)
Estimate a pattern spectrum with direct computation.
trains  (spike) train database to analyze
        The database must be an iterable of (spike) trains;
        each train must be a pair, consisting of a item/neuron
        identifier and an iterable of points/spike times;
        each item/neuron identifier must be a hashable object.
target  type of frequent item sets to find           (default: s)
        a/s  all/sets   frequent item sets
        as   seqs       frequent item sequences
width   window width/maximal distance between spikes (default: 0.003)
supp    minimum support/number of coincidences       (default: 2)
zmin    minimum item set size                        (default: 2)
zmax    maximum item set size                        (default: no limit)
report  pattern spectrum reporting format            (default: #)
        =    pattern spectrum as a list of triplets
        #    pattern spectrum as a dictionary
equiv   equivalent number of surrogate data sets     (default: 10000)
returns a pattern spectrum as a dictionary mapping pairs
        (size, support) to the corresponding occurrence counters
        or as a list of triplets (size, support, count)'''
    seq   = target.find('seq') >= 0 or target in ['as', 'cs', 'ms']
    equiv = float(equiv)        # get equivalent number of surrogates
    cnts  = countslots(trains, width)
    n     = len(trains)         # count slots per item set size
    if not zmax or zmax < 0 or len(cnts) < zmax: zmax = len(cnts)-1
    smax  = max(len(t) for i,t in trains)
    psp   = dict()              # initialize the pattern spectrum
    for z in range(zmin,zmax+1):# traverse the item set sizes
        k = float(choose(n,z))  # get number of possible item sets
        if seq: k *= factorial(z)
        l = cnts[z]/k           # and compute parameter lambda
        x = k *l *exp(-l)       # start of Poisson distribution
        for c in range(2,smax+1):
            y = l/c             # traverse the coincidences/support
            x *= y              # compute next distribution value
            e = int(x *equiv +0.5)
            if e <= 0:          # check value against threshold
                if y > 1.0: continue
                break           # stop if values only get smaller
            if c >= supp:       # add the computed signature
                psp[z,c] = e /equiv
    if report == '=':           # turn into a list of triplets
        psp = sorted([(z,c,psp[z,c]) for z,c in psp])
    return psp                  # return the estimated spectrum

#-----------------------------------------------------------------------

def estpsp_log (trains, target='s', width=0.003, supp=2,
                zmin=2, zmax=None, report='#', equiv=10000):
    '''estpsp_log (trains, target='s', width=0.003, supp=2,
            zmin=2, zmax=None, report='#', equiv=10000)
Estimate a pattern spectrum with logarithmic computation.
trains  (spike) train database to analyze
        The database must be an iterable of (spike) trains;
        each train must be a pair, consisting of a item/neuron
        identifier and an iterable of points/spike times;
        each item/neuron identifier must be a hashable object.
target  type of frequent item sets to find           (default: s)
        a/s  all/sets   frequent item sets
        as   seqs       frequent item sequences
width   window width/maximal distance between spikes (default: 0.003)
supp    minimum support/number of coincidences       (default: 2)
zmin    minimum item set size                        (default: 2)
zmax    maximum item set size                        (default: no limit)
report  pattern spectrum reporting format            (default: #)
        =    pattern spectrum as a list of triplets
        #    pattern spectrum as a dictionary
equiv   equivalent number of surrogate data sets     (default: 10000)
returns a pattern spectrum as a dictionary mapping pairs
        (size, support) to the corresponding occurrence counters
        or as a list of triplets (size, support, count)'''
    seq   = target.find('seq') >= 0 or target in ['as', 'cs', 'ms']
    equiv = float(equiv)        # get equivalent number of surrogates
    cnts  = countslots(trains, width)
    n     = len(trains)         # count slots per item set size
    if not zmax or zmax < 0 or len(cnts) < zmax: zmax = len(cnts)-1
    smax  = max(len(t) for i,t in trains)
    psp   = dict()              # initialize the pattern spectrum
    for z in range(zmin,zmax+1):# traverse the item set sizes
        y = lgamma(n+1) -lgamma(n-z+1)
        if not seq: y -= lgamma(z+1)
        l = log(cnts[z]) -y     # compute parameter lambda
        x = y -exp(l)           # start of Poisson distribution
        for c in range(1,smax+1):
            y = l -log(c)       # traverse the coincidences/support
            x += y              # compute next distribution value
            e = int(exp(x) *equiv +0.5)
            if e <= 0:          # check value against threshold
                if y > 0.0: continue
                break           # stop if values only get smaller
            if c >= supp:       # add the computed signature
                psp[z,c] = e /equiv
    if report == '=':           # turn into a list of triplets
        psp = sorted([(z,c,psp[z,c]) for z,c in psp])
    return psp                  # return the estimated spectrum

#-----------------------------------------------------------------------

def samplelp (probs, k):
    '''samplelp (probs, k)
Estimate the logarithm of the probability of a sample of k items.
probs   list of item probabilities
k       size of the sample to draw
returns the logarithm of the (estimated) sample probability.'''
    if k <= 0: return 0.0       # check for a valid argument
    s = sample(probs, k)        # draw a sample of size n from the items
    if k <= 1: return log(s[0]) # handle trivial case of size 1
    if k <= 2:                  # if two items have been selected
        p1,p2 = s; p = p1*p2    # compute probability explicitly
        p = p/(1.0-p1) +p/(1.0-p2)
        return log(p)           # return logarithm of probability
    if k <= 3:                  # if three items have been selected
        p1,p2,p3 = s; p = p1*p2*p3
        p = (p/(1.0-p2) +p/(1.0-p3)) /(1.0-p2-p3) \
          + (p/(1.0-p3) +p/(1.0-p1)) /(1.0-p1-p3) \
          + (p/(1.0-p1) +p/(1.0-p2)) /(1.0-p1-p2)
        return log(p)           # return logarithm of probability
    if k <= 4:                  # if four items have been selected
        p1,p2,p3,p4 = s; p = p1*p2*p3*p4
        p = ((p/(1.0-p3) +p/(1.0-p4)) /(1.0-p3-p4) \
          +  (p/(1.0-p4) +p/(1.0-p2)) /(1.0-p2-p4) \
          +  (p/(1.0-p2) +p/(1.0-p3)) /(1.0-p2-p3)) /(1.0-p2-p3-p4) \
          + ((p/(1.0-p3) +p/(1.0-p4)) /(1.0-p3-p4) \
          +  (p/(1.0-p4) +p/(1.0-p1)) /(1.0-p1-p4) \
          +  (p/(1.0-p1) +p/(1.0-p3)) /(1.0-p1-p3)) /(1.0-p1-p3-p4) \
          + ((p/(1.0-p2) +p/(1.0-p4)) /(1.0-p2-p4) \
          +  (p/(1.0-p4) +p/(1.0-p1)) /(1.0-p1-p4) \
          +  (p/(1.0-p1) +p/(1.0-p2)) /(1.0-p1-p2)) /(1.0-p1-p2-p4) \
          + ((p/(1.0-p2) +p/(1.0-p3)) /(1.0-p2-p3) \
          +  (p/(1.0-p3) +p/(1.0-p1)) /(1.0-p1-p3) \
          +  (p/(1.0-p1) +p/(1.0-p2)) /(1.0-p1-p2)) /(1.0-p1-p2-p3)
        return log(p)           # return logarithm of probability
    n = len(probs)              # get the number of probabilities/items
    p = 0.0                     # initialize average probability
    for i in range(8):          # draw 8/16 samples (permutations)
        if i <= 0: s.sort()     # uses sorted list in first loop
        else: shuffle(s)        # and some random permutation later
        t = r = 1.0             # init. the probabilities
        for q in s: t *= n*q/r; r -= q
        p += t                  # sum the permutation probability
        s.reverse()             # reverse the permutation
        t = r = 1.0             # init. the probabilities
        for q in s: t *= n*q/r; r -= q
        p += t                  # sum the permutation probability
        # The factor n in each step, which is removed again with the
        # term -k*log(n), serves the purpose to prevent underflow.
    return log(p/16) -k*log(n) +lgamma(k+1)

#-----------------------------------------------------------------------
 
def estpsp (trains, target='s', width=0.003, supp=2,
            zmin=2, zmax=None, report='#', equiv=10000,
            alpha=0.5, smpls=1000, seed=0):
    '''estpsp (trains, width=0.003, supp=2, zmin=2, zmax=None, report='#',
        equiv=10000, alpha=0.5, smpls=1000, seed=0)
Estimate a pattern spectrum with logarithmic computation.
trains  (spike) train database to analyze
        The database must be an iterable of (spike) trains;
        each train must be a pair, consisting of a item/neuron
        identifier and an iterable of points/spike times;
        each item/neuron identifier must be a hashable object.
target  type of frequent item sets to find           (default: s)
        a/s  all/sets   frequent item sets
        as   seqs       frequent item sequences
width   window width/maximal distance between spikes (default: 0.003)
supp    minimum support/number of coincidences       (default: 2)
zmin    minimum item set size                        (default: 2)
zmax    maximum item set size                        (default: no limit)
report  pattern spectrum reporting format            (default: #)
        =    pattern spectrum as a list of triplets
        #    pattern spectrum as a dictionary
alpha   probability dispersion factor                (default: 0.5)
smpls   number of samples to draw per item set size  (default: 1000)
seed    seed for the random number generator         (default: 0)
        (seed = 0: use system time as a seed)
equiv   equivalent number of surrogate data sets     (default: 10000)
returns a pattern spectrum as a dictionary mapping pairs
        (size, support) to the corresponding occurrence counters
        or as a list of triplets (size, support, count)'''
    if alpha <= 0:              # check for simple estimation
        return estpsp_log(trains, target, width, supp,
                          zmin, zmax, report, equiv)
    if seed == 0: seed = time() # get seed for random numbers
    srand(seed)                 # seed the random number generator
    seq   = target.find('seq') >= 0 or target in ['as', 'cs', 'ms']
    equiv = float(equiv)        # get equivalent number of surrogates
    cnts  = countslots(trains, width)
    n     = len(trains)         # count slots per item set size
    if zmax < 0 or len(cnts) < zmax: zmax = len(cnts)-1
    smax  = max(len(t) for i,t in trains)
    x     = 1/float(sum([len(t) for i,t in trains]))
    probs = [x*len(t) for i,t in trains]
    if alpha != 1.0:            # if to contract rate distribution
        x     = 1.0/n           # multiply deviation from mean w/ alpha
        probs = [(p-x)*alpha +x for p in probs]
    psp = dict()                # initialize the pattern spectrum
    for z in range(zmin,zmax+1):# traverse the pattern size range
        psp[z,0] = 0.0          # initialize the pattern spectrum
        q = lgamma(z+1) if seq else 0
        y = lgamma(n+1) -q -lgamma(n-z+1)
        t = -log(equiv) -y -9   # compute threshold for distribution
        for i in range(smpls):  # draw a certain number of samples
            l = log(cnts[z]) +samplelp(probs,z) -q
            x = -exp(l)         # compute distribution parameter
            psp[z,0] += exp(x)  # sum first distribution values
            for c in range(1,smax+1):
                y = l -log(c)   # traverse the coincidences/support
                x += y          # compute next distribution value
                if x < t:       # check value against threshold
                    if y > 0.0: continue
                    break       # stop if values only get smaller
                s = (z,c)       # update the pattern spectrum
                if s in psp: psp[s] += exp(x)
                else:        psp[s]  = exp(x)
        x = float(choose(n,z)) /smpls
        r = 0                   # compute the scaling factor
        for c in [c for y,c in psp if y == z]:
            psp[z,c] *= x       # scale the pattern spectrum
            r += psp[z,c] *c    # sum the covered slots
        x = cnts[z]/r if r > 0 else 1.0
        for c in [c for y,c in psp if y == z]:
            psp[z,c] *= x       # normalize the number of slots
        for c in [c for y,c in psp if y == z]:
            s = (z,c)           # traverse the pattern signatures
            if c < supp: del psp[s]; continue
            e = int(psp[s] *equiv +0.5)
            if e <= 0: del psp[s]    # remove unnecessary entries
            else: psp[s] = e /equiv  # and scale the others
    if report == '=':           # turn into a list of triplets
        psp = sorted([(z,c,psp[z,c]) for z,c in psp])
    return psp                  # return the estimated spectrum

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'Estimate a pattern spectrum for a train set'
    version = 'version 1.7 (2014.10.12)         ' \
            + '(c) 2013-2014   Christian Borgelt'
    args    = []                # list of program arguments/options
    opts    = {'o': ['seq',   False],  # flag for item sequences
               'w': ['width', 0.003],  # width of window/region
               's': ['supp',  2    ],  # minimum support (absolute)
               'm': ['zmin',  2    ],  # minimum number of items
               'n': ['zmax', -1    ],  # maximum number of items
               'a': ['beg',  -oo   ],  # minimum of point/time span
               'z': ['end',   oo   ],  # minimum of point/time span
               'c': ['equiv', 10000],  # equiv. number of surrogates
               'e': ['alpha', 0.5  ],  # probability dispersion factor
               'i': ['smpls', 1000 ],  # number of item set samples
               'S': ['seed',  0    ],  # seed for random numbers
               'l': ['line',  False],  # one train per line
               'y': ['item',  True ],  # item in line, item first
               'r': ['recseps', '\\n' ],     # record  separators
               'f': ['fldseps', ' ,\\t' ],   # field   separators
               'b': ['blanks',  ' \\t\\r' ], # blank   characters
               'C': ['comment', '#' ] }      # comment characters

    if len(argv) <= 1:          # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: estpsp.py [options] infile outfile')
        print(desc); print(version)
        print('-o       item sequences (respect item order)    '
                      +'(default: sets)');
        print('-w#      width of time window/maximum distance  '
                      +'(default: %g)' % opts['w'])
        print('-s#      minimum support of an item set         '
                      +'(default: %d)' % opts['s'])
        print('-m#      minimum number of items per item set   '
                      +'(default: %d)' % opts['m'])
        print('-n#      maximum number of items per item set   '
                      +'(default: no limit)')
        print('-a#      start of point/time range              '
                      +'(default: auto)')
        print('-z#      end   of point/time range              '
                      +'(default: auto)')
        print('-c#      equivalent number of surrogates        '
                      +'(default: %d)' % opts['c'])
        print('-e#      probability dispersion factor          '
                      +'(default: %g)' % opts['e'])
        print('-i#      number of item set samples per size    '
                      +'(default: %d)' % opts['i'])
        print('-S#      seed for pseudo-random numbers         '
                      +'(default: %d)' % opts['S'])
        print('-l       one train per line (item and points)   '
                      +'(default: id/time pairs)')
        print('-y       no item (with -l) or item after point  '
                      +'(default: item first)')
        print('-r#      record  separators                     '
                      +'(default: "%s")' % opts['r'])
        print('-f#      field   separators                     '
                      +'(default: "%s")' % opts['f'])
        print('-b#      blank   characters                     '
                      +'(default: "%s")' % opts['b'])
        print('-C#      comment characters                     '
                      +'(default: "%s")' % opts['C'])
        print('infile   file to read trains from               '
                      +'[required]')
        print('outfile  file to write pattern spectrum to      '
                      +'[optional]')
        exit()                  # print usage message and abort

    stderr.write('estpsp.py')   # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    # --- evaluate program options and arguments ---
    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 2: args.append(a); continue
            else: error('too many arguments\n')
        if a[1] not in opts:    # check for an option
            error('unknown option: %s\n' % a[:2])
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 1: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    seq     = opts['seq']       # flag for item sequences
    width   = opts['width']     # time window width/maximum distance
    supp    = opts['supp']      # minimum support/number of coincidences
    zmin    = opts['zmin']      # minimum size of item sets
    zmax    = opts['zmax']      # maximum size of item sets
    beg     = opts['beg']       # beginning of allowed span of spikes
    end     = opts['end']       # end       of allowed span of spikes
    equiv   = opts['equiv']     # equivalent number of surrogates
    alpha   = opts['alpha']     # probability dispersion factor
    smpls   = opts['smpls']     # number of item set samples per size
    seed    = opts['seed']      # random seed (0: use time)
    tpr     = opts['line']      # flag for one train per record
    item    = opts['item']      # flag for item first
    recseps = opts['recseps']   # record  separators
    fldseps = opts['fldseps']   # field   separators
    blanks  = opts['blanks']    # blank   characters
    comment = opts['comment']   # comment characters
    if zmin <= 0: error('invalid minimum size %d\n'    % zmin)
    if supp <= 0: error('invalid minimum support %g\n' % supp)
    if width < 0: error('invalid window width %g\n'    % width)
    x = [recseps,fldseps,blanks,comment]
    if version_info[0] >= 3:    # decode ASCII escape sequences
        x = [bytes(s, 'utf-8').decode('unicode_escape') for s in x]
    else:                       # decode ASCII escape sequences
        x = [s.decode('string_escape') for  s in x]
    recseps,fldseps,blanks,comment = x

    # --- read the data set to analyze ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[0])
    trains = dict(); k = 1      # initialize the train dictionary
    with open(args[0], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if rec == None: break
            if not rec: continue  # read the next record
            if tpr:             # if one train per line/record
                if item: n = rec[0]; rec = rec[1:]
                else:    n = k+1# get the item/neuron identifier
                trains[n] = [x for x in [float(x) for x in line]
                             if x >= beg and x <= end]
                k += 1          # collect the trains
            else:               # if item/ point pairs
                if item: n,s = rec
                else:    s,n = rec
                s = float(s)    # collect points per item/neuron
                if s < beg or s > end: continue
                if n in trains: trains[n].append(s)
                else:           trains[n] = [s]
    trains = trains.items()     # convert to pairs (item, list of times)
    points = sum([len(s) for n,s in trains])
    stderr.write('[%d item(s), %d point(s)]' % (len(trains), points))
    stderr.write(' done [%.2fs].\n' % (time()-t))
    if beg <= -oo: beg = floor(min(min(t) for n,t in trains))
    if end >= +oo: end = ceil (max(max(t) for n,t in trains))
    trains = [(i,[s for s in t if s >= beg and s <= end])
              for i,t in trains]# filter the trains/point processes

    # --- estimate pattern spectrum ---
    t = time()                  # start timer, print log message
    stderr.write('estimating pattern spectrum ... ')
    psp = estpsp(trains, 'seqs' if seqs else 's', width, supp,
                 zmin, zmax, '=', equiv, alpha, smpls, seed)
    stderr.write('[%d signature(s)]' % len(psp))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- write pattern spectrum ---
    t = time()                  # start timer, print log message
    stderr.write('writing %s ... ' % args[1])
    with open(args[1], 'w') as out:
        for s in psp: out.write('%2d %2d % .16g\n' % s)
    stderr.write('[%d signature(s)]' % len(psp))
    stderr.write(' done [%.2fs].\n' % (time()-t))
