#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : esteval.py
# Contents: evaluate pattern spectrum estimation
# Author  : Christian Borgelt
# History : 2014.02.21 file created
#           2014.03.06 adapted to modified estpsp.py
#           2015.03.13 adapted to modified pycoco library
#-----------------------------------------------------------------------
from sys     import argv, stderr, path
from os.path import join
from time    import strftime
from random  import seed as srand
from math    import log, ceil
try:
    from coco   import estpsp
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from estpsp import estpsp

#-----------------------------------------------------------------------

def cmppsp (spsp, epsp, thresh=0.001, high=0.1):
    spsp  = dict([(s,x) for s,x in spsp.items() if x >= thresh])
    epsp  = dict([(s,x) for s,x in epsp.items() if x >= thresh])
    sigs  = set(spsp.keys() +epsp.keys())
    sizes = sorted([z for z,c in sigs])
    for z in sizes:             # traverse the sizes
        cond = dict([(s,x) for s,x in spsp.items() if s[0] == z])
        fmax = max([x for x in cond.values()]+[0])
        if fmax > 0:            # adapt values before maximum
            cmax = min([s[1] for s,x in cond.items() if x == fmax])
            for c in range(2,cmax):
                if (z,c) in spsp: spsp[z,c] = fmax
        cond = dict([(s,x) for s,x in epsp.items() if s[0] == z])
        fmax = max([x for x in cond.values()]+[0])
        if fmax > 0:            # adapt values before maximum
            cmax = min([s[1] for s,x in cond.items() if x == fmax])
            for c in range(2,cmax):
                if (z,c) in spsp: epsp[z,c] = fmax
    n = max([z for z,c in sigs])+1
    d = [[i,0,0.0,0.0] for i in range(n)]
    for s in sigs:              # traverse the signatures
        a = spsp[s] if s in spsp else 0.0
        if a > high:   a = high # clamp the counter (surrogates)
        if a < thresh: a = 0.5*thresh
        b = epsp[s] if s in epsp else 0.0
        if b > high:   b = high # clamp the counter (estimated)
        if b < thresh: b = 0.5*thresh
        if a >= high and b >= high: continue
        x = log(b) -log(a)      # compute the logarithmic difference
        z,c = s                 # get pattern size and support
        d[z][1] += 1            # count the signatures per size
        d[z][2] += x            # sum the logarithmic differences
        d[z][3] += abs(x)       # and their absolute values per size
    return [[i, c, x/c if c > 0 else c, y/c if c > 0 else y, y]
            for i,c,x,y in d]   # normalize to pattern signatures

#-----------------------------------------------------------------------

def estimate (n=100, p=3, r=20, v=1, t=1, w=3,
              equiv=10000, alpha=1.0, smpls=1000, high=0.1):
    name   = 'n%d_p%g_r%g_v%d_t%d' % (n,p,r,v,t)
    trains = dict()             # initialize the train dictionary
    with open(join('data', name+'.txt'), 'r') as inp:
        for line in inp:        # traverse the lines of the input file
            n,s = line.split()  # split into item and spike time
            s = float(s.strip())# collect points per item/neuron
            if n in trains: trains[n].append(s)
            else:           trains[n] = [s]
    trains = trains.items()     # convert to pairs (item, list of times)
    name  += '_w%d' % w         # add the window width to the name
    #stderr.write(strftime('%Y.%m.%d %H:%M:%S '))
    #stderr.write(name); stderr.write('\n')
    spsp = dict()               # initialize the dictionary
    with open(join('data', 's_'+name+'.txt'), 'r') as inp:
        for line in inp:        # read pattern signatures and counts
            z,c,y = line.split()
            spsp[int(z),int(c)] = float(y)
    epsp = estpsp(trains, 's', 0.001*w, 2, 2, -1, '#',
                  equiv, alpha, smpls)
    diff = cmppsp(spsp, epsp, 1.0/equiv, high)
                                # estimate pattern spectrum
    epsp = sorted([(z,c,epsp[z,c]) for z,c in epsp])
    with open(join('data', 'e_'+name+'.txt'), 'w') as out:
        for s in epsp: out.write('%d %d %.16g\n' % s)
    return (name,diff)          # return difference to surrogates

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'evaluate pattern spectrum estimation'
    version = 'version 1.2 (2014.05.14)         ' \
            + '(c) 2014   Christian Borgelt'
    args    = []                  # list of program options
    opts    = {'n': ['ncnt',      0 ],   # number of neurons
               'p': ['period',    0 ],   # period in seconds
               'r': ['rate',      0 ],   # firing rate
               'v': ['nvar',      0 ],   # variation over neurons
               't': ['tvar',      0 ],   # variation over time
               'w': ['width',     0 ],   # width of analysis window
               'c': ['equiv', 10000 ],   # equiv. number of surrogates
               'h': ['high',    0.1 ],   # threshold for frequency
               'i': ['smpls',  1000 ],   # number of item set samples
               'e': ['alpha',   0.5 ] }  # prob. dispersion factor

    if len(argv) <= 1:         # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: esteval.py [options] outfile')
        print(desc); print(version)
        print('-n#      number of neurons                '
                      +'(in [100,80,60,40,0])')
        print('-p#      period in seconds                '
                      +'(in [3,0])')
        print('-r#      firing rate                      '
                      +'(in [30,20,10,0])')
        print('-v#      variation over neurons           '
                      +'(in [3,2,1,0])')
        print('-t#      variation over time              '
                      +'(in [3,2,1,0])')
        print('-w#      width of analysis window         '
                      +'(in [6,5,4,3,0])')
        print('-c#      equivalent number of surrogates  '
                      +'(default: %d)' % opts['c'])
        print('-h#      maximum for signature frequency  '
                      +'(default: %g)' % opts['h'])
        print('-i#      number of samples per set size   '
                      +'(default: %d)' % opts['i'])
        print('-e#      probability dispersion factor    '
                      +'(default: %g)' % opts['e'])
        print('outfile  file to write results to         '
                      +'[required]')
        exit()                  # print usage message and abort

    stderr.write('esteval.py')  # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 1: args.append(a); continue
            else: print('too many fixed arguments'); exit()
        if a[1] not in opts:    # check for an option
            print('unknown option: '+a[:2]); exit()
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 1: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    ncnt   = opts['ncnt']       # number of neurons
    period = opts['period']     # period in seconds
    rate   = opts['rate']       # firing rate in Hz
    nvar   = opts['nvar']       # variation of neurons
    tvar   = opts['tvar']       # variation over time
    width  = opts['width']      # width of analysis window
    equiv  = opts['equiv']      # equiv. number of surrogates
    high   = opts['high']       # threshold for frequency
    smpls  = opts['smpls']      # number of samples per set size
    alpha  = opts['alpha']      # probability dispersion factor

    if ncnt   not in [100,80,60,40,0]:
        stderr.write('invalid number of neurons %d\n'      % ncnt)
    if period not in [3,0]:
        stderr.write('invalid period %d\n'                 % period)
    if rate   not in [30,20,10,0]:
        stderr.write('invalid firing rate %d\n'            % rate)
    if nvar   not in [3,2,1,0]:
        stderr.write('invalid variation over neurons %d\n' % nvar)
    if tvar   not in [3,2,1,0]:
        stderr.write('invalid variation over time %d\n'    % tvar)
    if width  not in [6,5,4,3,0]:
        stderr.write('invalid analysis window width %d\n'  % width)

    ncnt   = [100,80,60,40] if ncnt   == 0 else [ncnt]
    period = [2,3,4,5]      if period == 0 else [period]
    rate   = [30,20,10]     if rate   == 0 else [rate]
    nvar   = [3,2,1]        if nvar   == 0 else [nvar]
    tvar   = [3,2,1]        if tvar   == 0 else [tvar]
    width  = [6,5,4,3]      if width  == 0 else [width]

    with open(args[0], 'w') as out:
        diffs = []              # init. the collected differences
        for n in ncnt:          # number of neurons
            for p in period:    # length of trains (duration)
                for r in rate:  # average firing rate
                    for v in [3,2,1]:     # variation over neurons
                        for t in [3,2,1]: # and over time
                            for w in [6,5,4,3]: # window width (ms)
                                f,d = estimate(n, p, r, v, t, w,
                                        equiv, alpha, smpls, high)
                                e = sum([z for i,c,x,y,z in d]) \
                                  / sum([c for i,c,x,y,z in d])
                                d = sum([y for i,c,x,y,z in d]) \
                                  / float(len(d)-2)
                                diffs.append((f,alpha,d,e))
        for d in diffs: out.write('%s %4.2f %8.6f %8.6f\n' % d);
        n = float(len(diffs))       # write results to a file
        print('%4.2f: %3d %8.6f %8.6f' %
             (alpha, n, sum([d for f,a,d,e in diffs])/n,
                        sum([e for f,a,d,e in diffs])/n))
